"""
Helpers to be available in the expression evaluation environment.
"""

import typing as T

if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    from .. import schedule as schedule_mod
    from ..room import Room

import datetime
import inspect
import itertools


class HelperBase:
    """A base for helpers to be available in the evaluation environment."""

    # Name under which helper should be available in environment
    namespace = ""

    # Helpers are applied in ascending order
    order = 0

    def __init__(
        self, room: "Room", now: datetime.datetime, env: T.Dict[str, T.Any]
    ) -> None:
        self._room = room
        self._app = room.app
        self._now = now
        self._env = env

    def update_environment(self) -> None:
        """Adds this helper instance or its attributes to the environment, depending
        on namespace."""

        if self.namespace:
            self._env[self.namespace] = self
        else:
            base_member_names = dir(HelperBase)
            for name, member in inspect.getmembers(self):
                if not name.startswith("_") and name not in base_member_names:
                    self._env[name] = member


class BasicHelper(HelperBase):
    """Adds some basic helpers."""

    def __init__(self, *args: T.Any, **kwargs: T.Any) -> None:
        super().__init__(*args, **kwargs)

        self.app = self._app
        self.room = self._room
        self.room_name = self._room.name

        self.datetime = datetime
        self.now = self._now
        self.date = self._now.date()
        self.time = self._now.time()

        self.schedule_snippets = self._app.cfg["schedule_snippets"]

    @staticmethod
    def is_empty(iterable: T.Iterable) -> bool:
        """Returns whether the given iterable is empty.
        next() is used for testing the iterable. For iterators, this
        has the side effect of the first item being consumed, but it
        avoids generating all values just for decision about emptiness."""

        try:
            if isinstance(iterable, T.Iterator):
                next(iterable)
            else:
                next(iter(iterable))
        except StopIteration:
            return True
        return False

    @staticmethod
    def round_to_step(
        value: T.Union[float, int],
        step: T.Union[float, int],
        decimal_places: int = None,
    ) -> T.Union[float, int]:
        """Round the value to the nearest step and, optionally, the
        given number of decimal places.
        Examples:
            round_to_step(34, 25) == 25
            round_to_step(0.665, 0.2, 1) == 0.6"""

        value = step * round(value / step)
        if decimal_places is not None:
            value = round(value, decimal_places)
        return value


class CustomEnvironmentHelper(HelperBase):
    """Merges the environment configured using expression_environment in."""

    order = 1000

    def update_environment(self) -> None:
        """Executes the expression_environment script."""

        script = self._app.expression_environment_script
        if script is not None:
            self._room.log(
                "Executing the expression_environment script.", level="DEBUG"
            )
            exec(script, self._env)  # pylint: disable=exec-used


class StateHelper(HelperBase):
    """Various state-related helpers."""

    def filter_entities(
        self, entities: T.Union[str, T.List[str]] = None, **criteria: T.Any
    ) -> T.Iterable[str]:
        """From a given set of entities, this function yields only those
        with a state and/or attributes matching all given criteria.
        Entities may either be specified as a single string (full entity
        id or domain), a list of such strings, or as None, which means
        all entities found in Home Assistant.
        The attributes of each entity are checked against the given
        criteria. Only entities that fulfill all criteria are yielded."""

        def _add_state(entity: str) -> None:
            if "." in entity:
                states[entity] = self.state(entity)
            else:
                states.update(self.state(entity))

        states = {}  # type: T.Dict[str, T.Any]
        if isinstance(entities, list):
            for entity in entities:
                _add_state(entity)
        elif isinstance(entities, str):
            _add_state(entities)
        else:
            states.update(self.state())

        for entity, state in states.items():
            if not isinstance(state, dict):
                continue
            attributes = state.get("attributes", {})
            for attr, value in criteria.items():
                if state.get(attr) != value and attributes.get(attr) != value:
                    break
            else:
                yield entity

    def is_off(self, entity_id: str) -> bool:
        """Returns whether an entity'S state is "off" (case-insensitive)."""

        return str(self.state(entity_id)).lower() == "off"

    def is_on(self, entity_id: str) -> bool:
        """Returns whether an entity'S state is "on" (case-insensitive)."""

        return str(self.state(entity_id)).lower() == "on"

    def state(self, entity: str = None, attribute: str = None) -> T.Any:
        """A wrapper around self._app.get_state()."""

        if entity and "." in entity:
            watched_entities = itertools.chain(
                self._room.cfg["watched_entities"], self._app.cfg["watched_entities"]
            )
            _attribute = attribute or "state"
            for watched in watched_entities:
                if entity != watched["entity"]:
                    continue
                if (
                    _attribute in watched["attributes"]
                    or "all" in watched["attributes"]
                ):
                    break
            else:
                suggestion = [entity]
                if attribute:
                    suggestion.append(attribute)
                self._room.log(
                    "You query {} from an expression, but didn't specify it "
                    "in the watched_entities config.".format(
                        repr(":".join(suggestion))
                    ),
                    level="WARNING",
                )

        if entity:
            if attribute:
                return self._app.get_state(entity, attribute=attribute)
            return self._app.get_state(entity)
        return self._app.get_state()


class ScheduleHelper(HelperBase):
    """Helpers related to schedule evaluation."""

    namespace = "schedule"

    def evaluate(
        self, schedule: "schedule_mod.Schedule", when: datetime.datetime = None
    ) -> T.Any:
        """Evaluates the given schedule for the given point in time."""

        when = when or self._now
        return schedule.evaluate(self._room, when)

    def next_results(
        self,
        schedule: "schedule_mod.Schedule",
        start: datetime.datetime = None,
        end: datetime.datetime = None,
    ) -> T.Generator[
        T.Tuple[datetime.datetime, "schedule_mod.ScheduleEvaluationResultType"],
        None,
        None,
    ]:
        """Returns a generator that yields tuples of datetime objects
        and schedule evaluation results. At each of these datetimes,
        the scheduling result will change to the returned one.
        The first result generated is always that for the start time,
        the last one that for the end time."""

        when = start or self._now  # type: T.Optional[datetime.datetime]
        last_result = None
        while when and (not end or end > when):
            result = schedule.evaluate(self._room, when)
            if result and result != last_result:
                yield when, result
                last_result = result
            when = schedule.get_next_scheduling_datetime(when)

        if end and last_result:
            yield end, last_result


class PatternHelper(HelperBase):
    """Help generate values based on different patterns."""

    namespace = "pattern"

    @staticmethod
    def linear(
        start_value: T.Union[float, int],
        end_value: T.Union[float, int],
        percentage: T.Union[float, int],
    ) -> float:
        """Calculate the value at a given percentage between start_value
        and end_value."""

        return start_value + percentage * (end_value - start_value) / 100



"""
This module implements a binary on/off switch, derived from the generic actor.
"""

from .generic import GenericActor


class SwitchActor(GenericActor):
    """A binary on/off switch actor for Schedy."""

    name = "switch"
    config_defaults = {
        **GenericActor.config_defaults,
        "attributes": [
            {
                "attribute": "state",
                "values": {
                    "on": {"service": "homeassistant/turn_on"},
                    "off": {"service": "homeassistant/turn_off"},
                },
            }
        ],
    }



"""
This module implements the generic actor.
"""

import typing as T

import voluptuous as vol

from ... import common
from .base import ActorBase


ALLOWED_VALUE_TYPES = (float, int, str, type(None))
ALLOWED_VALUE_TYPES_T = T.Union[float, int, str, None]  # pylint: disable=invalid-name

VALUE_DEF_SCHEMA = vol.Schema(
    vol.All(
        lambda v: v or {},
        {
            vol.Required("service"): vol.All(str, lambda v: v.replace(".", "/", 1)),
            vol.Optional("service_data", default=dict): vol.All(
                lambda v: v or {}, dict
            ),
            vol.Optional("include_entity_id", default=True): bool,
            vol.Optional("value_parameter", default=None): vol.Any(str, None),
        },
    )
)

WILDCARD_VALUE_SCHEMA = vol.Schema(vol.All(vol.Coerce(str), str.lower, "_other_"))


class GenericActor(ActorBase):
    """A configurable, generic actor for Schedy that can control multiple
    attributes at once."""

    name = "generic"
    config_schema_dict = {
        **ActorBase.config_schema_dict,
        vol.Optional("attributes", default=list): vol.All(
            lambda v: v or [],
            [
                vol.All(
                    lambda v: v or {},
                    {
                        vol.Optional("attribute", default=None): vol.Any(str, None),
                        vol.Optional("values", default=dict): vol.All(
                            lambda v: v or {},
                            {
                                vol.Any(
                                    WILDCARD_VALUE_SCHEMA, *ALLOWED_VALUE_TYPES
                                ): VALUE_DEF_SCHEMA
                            },
                        ),
                    },
                )
            ],
        ),
        vol.Optional("call_reversed", default=False): bool,
        vol.Optional("short_values", default=list): vol.All(
            lambda v: v or [],
            [vol.Any(WILDCARD_VALUE_SCHEMA, *ALLOWED_VALUE_TYPES), vol.Coerce(tuple)],
            lambda v: sorted(v, key=len, reverse=True),
        ),
    }

    def _get_value_cfg(
        self, index: int, value: ALLOWED_VALUE_TYPES_T
    ) -> T.Tuple[ALLOWED_VALUE_TYPES_T, T.Dict]:
        """Returns the key and value configuration for given attribute
        index and value.
        A KeyError is raised when the value is unknown and no _other_
        is configured."""

        for key in (value, "_other_"):
            try:
                return key, self.cfg["attributes"][index]["values"][key]
            except KeyError:
                pass

        raise KeyError(value)

    def do_send(self) -> None:
        """Executes the services configured for self._wanted_value."""

        value = self._wanted_value
        if not isinstance(value, tuple):
            value = (value,)

        iterator = enumerate(value)  # type: T.Iterator[T.Tuple[int, T.Any]]
        if self.cfg["call_reversed"]:
            iterator = reversed(list(iterator))

        for index, item in iterator:
            _, cfg = self._get_value_cfg(index, item)
            service = cfg["service"]
            service_data = cfg["service_data"].copy()
            if cfg["include_entity_id"]:
                service_data.setdefault("entity_id", self.entity_id)
            if cfg["value_parameter"] is not None:
                service_data.setdefault(cfg["value_parameter"], item)

            self.log(
                "Calling service {}, data = {}.".format(
                    repr(service), repr(service_data)
                ),
                level="DEBUG",
                prefix=common.LOG_PREFIX_OUTGOING,
            )
            self.app.call_service(service, **service_data)

    def filter_set_value(self, value: T.Any) -> T.Any:
        """Checks whether the actor supports this value."""

        def _log_invalid_length() -> None:
            self.log(
                "The value {} has not the expected number of items "
                "({} (actual) vs. {} (expected)).".format(
                    repr(value), len(value), len(self.cfg["attributes"])
                ),
                level="ERROR",
            )

        if not isinstance(value, tuple):
            value = (value,)
        if len(value) > len(self.cfg["attributes"]):
            _log_invalid_length()
            return None

        _value = []
        for index, item in enumerate(value):
            try:
                self._get_value_cfg(index, item)
            except KeyError:
                self.log(
                    "Value {} for slot {} is not known by this "
                    "generic actor, ignoring request to set it.".format(
                        repr(item), index
                    ),
                    level="ERROR",
                )
                return None
            _value.append(item)

        for short in self.cfg["short_values"]:
            if short == tuple(_value[: len(short)]):
                if len(_value) > len(short):
                    self.log(
                        "VALUe {} should be shortened to {}, "
                        "doing it for you now.".format(repr(_value), repr(short)),
                        level="WARNING",
                    )
                    _value = short
                    return None
                break
        else:
            if len(value) != len(self.cfg["attributes"]):
                _log_invalid_length()
                return None

        return value[0] if len(_value) == 1 else tuple(_value)

    def notify_state_changed(self, attrs: dict) -> T.Any:
        """Is called when the entity's state changes."""

        items = []  # type: T.List[T.Any]
        path = []  # type: T.List[ALLOWED_VALUE_TYPES_T]
        for index, cfg in enumerate(self.cfg["attributes"]):
            attr = cfg["attribute"]
            if attr is None:
                # write-only slot, can never be determined
                items.append(None)
                path.append(None)
                continue

            state = attrs.get(attr)
            self.log(
                "Attribute {} is {}.".format(repr(attr), repr(state)),
                level="DEBUG",
                prefix=common.LOG_PREFIX_INCOMING,
            )
            if state is None:
                self.log("Ignoring incomplete state change.", level="DEBUG")
                return None

            try:
                key, _ = self._get_value_cfg(index, state)
            except KeyError:
                self.log(
                    "State {} for slot {} is not known by this "
                    "generic actor, ignoring state change.".format(repr(state), index),
                    level="ERROR",
                )
                return None

            items.append(state)
            path.append(key)

            for short in self.cfg["short_values"]:
                if short == tuple(path):
                    self.log(
                        "This is a configured short value {}, breaking.".format(
                            repr(short)
                        ),
                        level="DEBUG",
                    )
                    break
            else:
                continue
            break

        return items[0] if len(items) == 1 else tuple(items)

    @staticmethod
    def validate_value(value: T.Any) -> T.Any:
        """Converts lists to tuples."""

        if isinstance(value, list):
            items = tuple(value)
        elif isinstance(value, tuple):
            items = value
        else:
            items = (value,)

        for index, item in enumerate(items):
            if not isinstance(item, ALLOWED_VALUE_TYPES):
                raise ValueError(
                    "Value {} for slot {} must be of one of these types: {}".format(
                        repr(item), index, ALLOWED_VALUE_TYPES
                    )
                )

        return items[0] if len(items) == 1 else items



"""
This module contains code that dynamically scans for available apps
at module load time. For all apps found, a loader is generated which,
when called, imports the particular app and behaves like the app
class itself. The loaders are made available as module attributes
under the same name the corresponding app classes would have
(e.g. SchedyApp).
The __all__ list is populated with these loaders, hence a wildcard
import will fetch them all.
"""

import typing as T
import types

import importlib
import os
import sys


def _import_app_module(package: str) -> types.ModuleType:
    mod_name = "{}.app".format(package)
    if __package__:
        mod_name = "{}.{}".format(__package__, mod_name)
    return importlib.import_module(mod_name)


def _build_app_loader(app_package: str, app_class_name: str) -> T.Callable:
    def _proxy_loader(*args, **kwargs):  # type: ignore
        app_mod = _import_app_module(app_package)
        app_class = getattr(app_mod, app_class_name)
        return app_class(*args, **kwargs)

    return _proxy_loader


def _generate_app_loaders() -> T.Iterable[T.Tuple[str, T.Callable]]:
    """Scans for apps and yields tuples of the app class name and a
    deferred loader for each app found."""

    dirpath = os.path.realpath(os.path.dirname(__file__))
    for name in os.listdir(dirpath):
        path = os.path.join(dirpath, name)
        if not os.path.isdir(path) or not os.path.isfile(os.path.join(path, "app.py")):
            continue
        parts = [part.capitalize() for part in name.split("_")]
        attr = "{}App".format("".join(parts))
        loader = _build_app_loader(name, attr)
        yield attr, loader


# make app loaders available as module attributes
__all__ = []
for _attr, _loader in _generate_app_loaders():
    setattr(sys.modules[__name__], _attr, _loader)
    __all__.append(_attr)



"""
Utility functions that are used all around Schedy.
"""

import types
import typing as T

import collections
import datetime
import re
import voluptuous as vol


# matches any character not allowed in Python variable names
INVALID_VAR_NAME_CHAR_PATTERN = re.compile(r"[^0-9a-z_]", re.I)

# regexp pattern matching a range spec like 5, 3-7 or */5 without spaces
RANGE_PATTERN = re.compile(r"^(?:(\*)|(\d+)(?:\-(\d+))?)(?:\/([1-9]\d*))?$")

# strftime-compatible format string for military time
TIME_FORMAT = "%H:%M:%S"
# regular expression for time formats
TIME_REGEXP = re.compile(
    r"^ *(?P<h>[01]?\d|2[0-3]) *\: *(?P<m>[0-5]\d) *(?:\: *(?P<s>[0-5]\d) *)?$"
)

# Regular expression for rule start/end times with day shift
RULE_TIME_REGEXP = re.compile(
    r"^(?P<time> *(?P<h>[01]?\d|2[0-3]) *\: *(?P<m>[0-5]\d) *"
    r"(?:\: *(?P<s>[0-5]\d) *)?)?(?:(?P<days>[+-] *\d+) *d)?$"
)

# used instead of vol.Extra to ensure keys are strings
CONF_STR_KEY = vol.Coerce(str)


class RangingSet(set):
    """A set for integers that forms nice ranges in its __repr__,
    perfectly suited for the expansion of range strings."""

    def __repr__(self) -> str:
        if not self:
            return "{}"

        # fall back to legacy representation when non-ints are found
        for item in self:
            if not isinstance(item, int):
                return super().__repr__()

        nums = sorted(self)  # type: T.List[int]
        ranges = collections.OrderedDict()  # type: T.Dict[int, int]
        range_start = nums[0]
        ranges[range_start] = range_start
        for num in nums[1:]:
            if num - 1 != ranges[range_start]:
                range_start = num
            ranges[range_start] = num

        return "{{{}}}".format(
            ", ".join(
                [
                    str(start) if start == end else "{}-{}".format(start, end)
                    for start, end in ranges.items()
                ]
            )
        )


def build_date_from_constraint(
    constraint: T.Dict[str, int], default_date: datetime.date, direction: int = 0
) -> datetime.date:
    """Builds and returns a datetime.date object from the given constraint,
    taking missing values from the given default_date.
    In case the date is not valid (e.g. 2017-02-29), a ValueError is
    raised, unless a number has been given for direction, in which case
    the next/previous valid date will be chosen, depending on the sign
    of direction."""

    fields = {}
    for field in ("year", "month", "day"):
        fields[field] = constraint.get(field, getattr(default_date, field))

    while True:
        try:
            return datetime.date(**fields)
        except ValueError:
            if direction > 0:
                fields["day"] += 1
            elif direction < 0:
                fields["day"] -= 1
            else:
                raise

            # handle month/year transitions correctly
            if fields["day"] < 1:
                fields["day"] = 31
                fields["month"] -= 1
            elif fields["day"] > 31:
                fields["day"] = 1
                fields["month"] += 1
            if fields["month"] < 1:
                fields["month"] = 12
                fields["year"] -= 1
            elif fields["month"] > 12:
                fields["month"] = 1
                fields["year"] += 1


def compile_expression(expr: str) -> types.CodeType:
    """Compiles strings to code objects.
    Strings containing one or more newlines are assumed to contain
    whole statements. Others are treated as simple expressions and
    "result = " is prepended to them before compilation is done as a
    single statement."""

    if "\n" in expr:
        mode = "exec"
    else:
        expr = "result = {}".format(expr)
        mode = "single"

    compiled = compile(
        expr, "expression", mode, dont_inherit=True
    )  # type: types.CodeType
    return compiled


def deep_merge_dicts(source: dict, dest: dict) -> None:
    """Updates items of dest with those of source, descending into and
    merging child dictionaries as well. Child lists are combined as
    well so that the items of a list from source are appended to the
    list found in dest under the same name."""

    for key, value in source.items():
        dest_value = dest.get(key)
        if isinstance(value, dict) and isinstance(dest_value, dict):
            deep_merge_dicts(value, dest[key])
        elif isinstance(value, list) and isinstance(dest_value, list):
            dest_value.extend(value)
        else:
            dest[key] = value


def escape_var_name(name: str) -> str:
    """Converts the given string to a valid Python variable name.
    All unsupported characters are replaced by "_". If name would
    start with a digit, "_" is put infront."""

    name = INVALID_VAR_NAME_CHAR_PATTERN.sub("_", name)
    digits = tuple([str(i) for i in range(10)])
    if name.startswith(digits):
        name = "_" + name
    return name


def expand_range_spec(
    spec: T.Union[int, str], min_value: int, max_value: int
) -> T.Set[int]:
    """Expands strings of the range specification format to RangingSet
    objects containing the individual numbers.
    Any whitespace is ignored. If an int is given instead of a string,
    a set containing only that is returned.
    The min_value and max_value are required to support the * specifier
    and a ValueError is raised when the range specification exceeds
    these boundaries."""

    if isinstance(spec, int):
        spec = str(spec)
    spec = "".join(spec.split())
    if spec.startswith("!"):
        spec = spec[1:]
        invert = True
    else:
        invert = False

    numbers = RangingSet()
    for part in spec.split(","):
        match = RANGE_PATTERN.match(part)
        if match is None:
            raise ValueError("invalid range definition: {}".format(repr(part)))

        _wildcard, _start, _end, _step = match.groups()
        if _wildcard:
            start = min_value
            end = max_value
        else:
            start = int(_start)
            end = start if _end is None else int(_end)
            for value in (start, end):
                if value < min_value or value > max_value:
                    raise ValueError(
                        "value {} is out of range {}..{}".format(
                            value, min_value, max_value
                        )
                    )
            if end < start:
                start, end = end, start
        step = int(_step or 1)

        for i in range(start, end + 1, step):
            numbers.add(i)

    if invert:
        numbers.symmetric_difference_update(range(min_value, max_value + 1))

    return numbers


def format_time(when: datetime.time, format_str: str = TIME_FORMAT) -> str:
    """Returns a string representing the given datetime.time object.
    If no strftime-compatible format is provided, the default is used."""

    return when.strftime(format_str)


def normalize_dict_key(
    obj: dict, dest_key: T.Any, *alt_keys: T.Any, keep_alt_keys: bool = False
) -> None:
    """If dest_key is missing in the dict obj but one of the alt_keys
    is found instead, the value of the first alt_key is moved over
    to dest_key.
    All alt_keys are deleted from dict obj unless keep_alt_keys is True."""

    for alt_key in alt_keys:
        if alt_key in obj:
            if dest_key not in obj:
                obj[dest_key] = obj[alt_key]
            if not keep_alt_keys:
                del obj[alt_key]


def parse_time_string(time_str: str) -> datetime.time:
    """Parses a string recognizable by TIME_REGEXP into a datetime.time object. If
    the string has an invalid format, a ValueError is raised."""

    match = TIME_REGEXP.match(time_str)
    if match is None:
        raise ValueError("time string {} has an invalid format".format(repr(time_str)))
    groups = match.groupdict()
    return datetime.time(int(groups["h"]), int(groups["m"]), int(groups["s"] or 0))


def parse_rule_time_string(
    time_str: str,
) -> T.Tuple[T.Optional[datetime.time], T.Optional[int]]:
    """Parses a string recognizable by RULE_TIME_REGEXP into a (datetime.time,
    int) tuple. Both time and days are optional and None if not specified. If the
    string has an invalid format, a ValueError is raised."""

    match = RULE_TIME_REGEXP.match(time_str)
    if match is None:
        raise ValueError("time string {} has invalid format".format(repr(time_str)))
    groups = match.groupdict()
    _time = None
    if groups["time"]:
        _time = datetime.time(int(groups["h"]), int(groups["m"]), int(groups["s"] or 0))
    days = None if groups["days"] is None else int(groups["days"])
    return _time, days


def round_number(number: T.Union[float, int], places: int) -> T.Union[float, int]:
    """Rounds the given number to the given decimal places. If places
    is 0, an integer is returned, a float otherwise."""

    if places < 0:
        raise ValueError("can'T round to a negative number of decimal places")

    if places == 0:
        return round(number, 0)

    string = "{{:.{}f}}".format(places).format(number).rstrip("0")
    if string.endswith("."):
        string += "0"
    return float(string)



"""
This module implements the Schedule and Rule classes.
"""

import typing as T

if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    import types
    from .room import Room

import datetime
import functools

from cached_property import cached_property

from . import util
from . import expression


ScheduleEvaluationResultType = T.Tuple[T.Any, T.Set[str], "Rule"]


class Rule:
    """A rule that can be added to a schedule."""

    # names of schedule rule constraints to be fetched from a rule definition
    CONSTRAINTS = (
        "years",
        "months",
        "days",
        "weeks",
        "weekdays",
        "start_date",
        "end_date",
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str = None,
        start_time: datetime.time = None,
        start_plus_days: int = None,
        end_time: datetime.time = None,
        end_plus_days: int = None,
        constraints: T.Dict[str, T.Any] = None,
        expr: "types.CodeType" = None,
        expr_raw: str = None,
        value: T.Any = None,
    ) -> None:
        _checks = [expr is None, expr_raw is None]
        if any(_checks) and not all(_checks):
            raise ValueError("expr and expr_raw may only be passed together")
        if expr is not None and value is not None:
            raise ValueError("specify only one of expr and value, not both")

        self.name = name

        self.start_time = start_time
        self.start_plus_days = start_plus_days
        self.end_time = end_time
        self.end_plus_days = end_plus_days

        if constraints is None:
            constraints = {}
        self.constraints = constraints

        self.expr = expr
        self.expr_raw = expr_raw
        self.value = value

        # We cache constraint check results for the latest-checked 64 days
        self.check_constraints = functools.lru_cache(maxsize=64)(
            self._check_constraints
        )

    def __repr__(self) -> str:
        return "<Rule {}{}>".format(
            "{} ".format(repr(self.name)) if self.name is not None else "",
            ", ".join(self._get_repr_tokens()),
        )

    @staticmethod
    def _format_constraint(data: T.Any) -> str:
        """Formats constraint values for use in __repr__."""

        return str(data).replace(" ", "").replace("'", "")

    @staticmethod
    def _format_time(_time: datetime.time = None, days: int = None) -> str:
        """Formats time + shift days as a string for use in __repr__."""

        if _time is None:
            time_repr = "??:??"
        else:
            time_repr = _time.strftime("%H:%M:%S" if _time.second else "%H:%M")
        if days is None:
            days_repr = ""
        elif days < 0:
            days_repr = "{}d".format(days)
        else:
            days_repr = "+{}d".format(days)
        return "{}{}".format(time_repr, days_repr)

    def _get_repr_tokens(self) -> T.List[str]:
        """Returns a list of tokens to be shown in repr()."""

        tokens = []  # type: T.List[str]

        if (
            self.start_time is not None
            or self.start_plus_days is not None
            or self.end_time is not None
            or self.end_plus_days is not None
        ):
            times = "from {} to {}".format(
                self._format_time(self.start_time, self.start_plus_days),
                self._format_time(self.end_time, self.end_plus_days),
            )
            tokens.append(times)

        for constraint in sorted(self.constraints):
            tokens.append(
                "{}={}".format(
                    constraint, self._format_constraint(self.constraints[constraint])
                )
            )

        if self.expr_raw is not None:
            if len(self.expr_raw) > 43:
                tokens.append("x={}...".format(repr(self.expr_raw[:40])))
            else:
                tokens.append("x={}".format(repr(self.expr_raw)))

        if self.value is not None:
            tokens.append("v={}".format(repr(self.value)))

        return tokens

    def _check_constraints(self, date: datetime.date) -> bool:
        """Checks all constraints of this rule against the given date
        and returns whether they are fulfilled"""

        if not self.constraints:
            return True

        year, week, weekday = date.isocalendar()
        checks = {
            "years": lambda a: year in a,
            "months": lambda a: date.month in a,
            "days": lambda a: date.day in a,
            "weeks": lambda a: week in a,
            "weekdays": lambda a: weekday in a,
            "start_date": lambda a: date >= util.build_date_from_constraint(a, date, 1),
            "end_date": lambda a: date <= util.build_date_from_constraint(a, date, -1),
        }

        for constraint, allowed in self.constraints.items():
            if not checks[constraint](allowed):  # type: ignore
                return False
        return True


class RulePath:
    """A chain of rules starting from a root schedule through sub-schedule
    rules."""

    def __init__(self, root_schedule: "Schedule") -> None:
        self.root_schedule = root_schedule
        self.rules = []  # type: T.List[Rule]

    def __add__(self, other: "RulePath") -> "RulePath":
        """Creates a new RulePath with rules of self and another path.
        The paths have to fit together, meaning the root schedule of the other path
        must be the sub schedule of the rightmost rule of this path."""

        if (
            not isinstance(other, RulePath)
            or not self.rules
            or not isinstance(self.rules[-1], SubScheduleRule)
            or self.rules[-1].sub_schedule is not other.root_schedule
        ):
            raise ValueError("{!r} and {!r} don't fit together".format(self, other))

        path = self.copy()
        path.extend(other.rules)
        return path

    def __repr__(self) -> str:
        if not self.rules:
            return "<{}/empty rule path>".format(self.root_schedule)

        locs = []
        sched = self.root_schedule
        for rule in self.rules:
            if rule in sched.rules:
                loc = str(sched.rules.index(rule) + 1)
            else:
                loc = "?"
            locs.append(loc)
            if not isinstance(rule, SubScheduleRule):
                break
            sched = rule.sub_schedule

        return "<{}/{}:{}>".format(
            self.root_schedule,
            "/".join(locs),
            rule,  # pylint: disable=undefined-loop-variable
        )

    def _clear_cache(self) -> None:
        """Clears out all cached properties. For internal use only."""

        for attr in ("is_always_active", "rules_with_expr_or_value", "times"):
            try:
                del self.__dict__[attr]
            except KeyError:
                pass

    def append(self, rule: Rule, clear_cache: bool = True) -> None:
        """Add's a rule to the end of the path.
        A ValueError is raised when the previous rule is a final rule."""

        if self.rules and not isinstance(self.rules[-1], SubScheduleRule):
            raise ValueError(
                "The previous rule in the path ({}) is no SubScheduleRule.".format(
                    self.rules[-1]
                )
            )

        self.rules.append(rule)

        if clear_cache:
            self._clear_cache()

    def copy(self) -> "RulePath":
        """Returns a mutable copy of this path."""

        path = type(self)(self.root_schedule)
        path.extend(self.rules)
        return path

    def extend(self, rules: T.Iterable[Rule]) -> None:
        """Appends each of the supplied rules to the path.
        If append() raises a ValueError for one of the rules to add, all rules already
        appended successfully are removed again before the exception is re-raised.
        Note: This method is not thread-safe!"""

        added = 0
        try:
            for rule in rules:
                self.append(rule, clear_cache=False)
                added += 1
        except ValueError:
            for _ in range(added):
                self.pop(clear_cache=False)
            added = 0
            raise
        finally:
            if added:
                self._clear_cache()

    def includes_schedule(self, schedule: "Schedule") -> bool:
        """Checks whether the given schedule is included in this path."""
        if schedule is self.root_schedule:
            return True
        for rule in self.rules:
            if isinstance(rule, SubScheduleRule) and rule.sub_schedule is schedule:
                return True
        return False

    def check_constraints(self, date: datetime.date) -> bool:
        """Checks constraints of all rules along this path against the
        given date and returns whether they are all fulfilled."""

        for rule in self.rules:
            if not rule.check_constraints(date):
                return False
        return True

    def is_active(self, when: datetime.datetime) -> bool:
        """Returns whether the rule this path leads to is active at
        given point in time."""

        # Short-circuit the algorithm if possible
        if self.is_always_active:
            return True

        _date, _time = when.date(), when.time()
        start_time, start_plus_days, end_time, end_plus_days = self.times

        # We first build a list of possible dates on which the path could start
        # being active, then we check whether one of these dates fulfills the path's
        # constraints.
        shift_list = list(range(end_plus_days + 1))
        # List isn't empty, 0 will always be included
        if _time >= end_time:
            # Don't check the most distant date, already ended earlier today
            del shift_list[-1]
        if shift_list and _time < start_time:
            # Today can't be the starting day, it's too early
            del shift_list[0]

        # Now check the constraints for each possible starting date
        for days_back in shift_list:
            start_date = _date - datetime.timedelta(days=days_back + start_plus_days)
            if not self.check_constraints(start_date):
                # Not starting this day, try next one
                continue
            # Found valid starting day, path is active now
            return True
        return False

    @cached_property
    def is_always_active(self) -> bool:
        """Whether this path is always active (no constraints and spans >= 1 day)."""

        for rule in self.rules:
            if rule.constraints:
                return False

        start_time, _, end_time, end_plus_days = self.times
        return bool(end_plus_days > 1 or end_plus_days == 1 and end_time >= start_time)

    @property
    def is_final(self) -> bool:
        """Returns whether the last rule in the path is no SubScheduleRule."""

        if not self.rules:
            return False
        return not isinstance(self.rules[-1], SubScheduleRule)

    def pop(self, clear_cache: bool = True) -> Rule:
        """Removes and returns the rightmost rule of this path.
        IndexError is raised when there is no rule to pop."""

        try:
            rule = self.rules.pop()
        except IndexError:
            raise IndexError("no rule to pop")
        if clear_cache:
            self._clear_cache()
        return rule

    @cached_property
    def rules_with_expr_or_value(self) -> T.Tuple[Rule, ...]:
        """A tuple with rules of the path containing an expression or value,
        sorted from left to right."""

        return tuple(
            filter(lambda r: r.expr is not None or r.value is not None, self.rules)
        )

    @cached_property
    def times(self) -> T.Tuple[datetime.time, int, datetime.time, int]:
        """Returns (start_time, start_plus_days, end_time, end_plus_days) for this
        path. Rules are searched for these values from right to left.
        Missing times are assumed to be midnight. If not set explicitly, end_plus_days
        is 1 if start <= end else 0."""

        for rule in reversed(self.rules):
            if rule.start_time is not None:
                start_time = rule.start_time
                break
        else:
            start_time = datetime.time(0, 0)

        for rule in reversed(self.rules):
            if rule.start_plus_days is not None:
                start_plus_days = rule.start_plus_days
                break
        else:
            start_plus_days = 0

        for rule in reversed(self.rules):
            if rule.end_time is not None:
                end_time = rule.end_time
                break
        else:
            end_time = datetime.time(0, 0)

        for rule in reversed(self.rules):
            if rule.end_plus_days is not None:
                end_plus_days = rule.end_plus_days
                break
        else:
            end_plus_days = 1 if start_time >= end_time else 0

        return start_time, start_plus_days, end_time, end_plus_days


class SubScheduleRule(Rule):
    """A schedule rule with a sub-schedule attached."""

    def __init__(self, sub_schedule: "Schedule", *args: T.Any, **kwargs: T.Any) -> None:
        super().__init__(*args, **kwargs)

        self.sub_schedule = sub_schedule

    def _get_repr_tokens(self) -> T.List[str]:
        """Adds the sub-schedule information to repr()."""

        tokens = super()._get_repr_tokens()
        tokens.insert(0, "with sub {}".format(self.sub_schedule))
        return tokens


class Schedule:
    """Holds the schedule for a room with all its rules."""

    def __init__(self, name: str = None, rules: T.Iterable[Rule] = None) -> None:
        self.name = name
        self.rules = []  # type: T.List[Rule]
        if rules is not None:
            self.rules.extend(rules)

    def __add__(self, other: "Schedule") -> "Schedule":
        if not isinstance(other, Schedule):
            return NotImplemented
        return Schedule(name=self.name, rules=self.rules + other.rules)

    def __repr__(self) -> str:
        if self.name is None:
            return "<Schedule of {} rules>".format(len(self.rules))
        return "<Schedule {}>".format(repr(self.name))

    def evaluate(  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        self, room: "Room", when: datetime.datetime
    ) -> T.Optional[ScheduleEvaluationResultType]:
        """Evaluates the schedule, computing the value for the time the
        given datetime object represents. The resulting value, a set of
        markers applied to the value and the matching rule are returned.
        If no value could be found in the schedule (e.g. all rules
        evaluate to Next()), None is returned."""

        def log(msg: str, path: RulePath, *args: T.Any, **kwargs: T.Any) -> None:
            """Wrapper around room.log that prefixes spaces to the
            message based on the length of the rule path."""

            prefix = " " * 4 * max(0, len(path.rules) - 1) + "\u251c\u2500"
            room.log("{} {}".format(prefix, msg), *args, **kwargs)

        room.log("Assuming it to be {}.".format(when), level="DEBUG")

        expr_cache = {}  # type: T.Dict[types.CodeType, T.Any]
        expr_env = None
        markers = set()  # type: T.Set[str]
        postprocessors = []
        paths = list(self.unfolded)
        path_idx = 0
        while path_idx < len(paths):
            path = paths[path_idx]
            path_idx += 1

            last_rule = path.rules[-1]
            if isinstance(last_rule, SubScheduleRule):
                log("[SUB]  {}".format(path), path, level="DEBUG")
                continue
            if not path.is_active(when):
                log("[INA]  {}".format(path), path, level="DEBUG")
                continue
            log("[ACT]  {}".format(path), path, level="DEBUG")

            result = None
            for rule in reversed(path.rules_with_expr_or_value):
                if rule.expr is not None:
                    plain_value = False
                    try:
                        result = expr_cache[rule.expr]
                    except KeyError:
                        if expr_env is None:
                            expr_env = expression.build_expr_env(room, when)
                        result = room.eval_expr(rule.expr, expr_env)
                        expr_cache[rule.expr] = result
                        log("=> {}".format(repr(result)), path, level="DEBUG")
                        # Unwrap a result with markers
                        if isinstance(result, expression.types.Mark):
                            result = result.unwrap(markers)
                    else:
                        log(
                            "=> {}  [cache-hit]".format(repr(result)),
                            path,
                            level="DEBUG",
                        )
                    if isinstance(result, Exception):
                        room.log(
                            "Failed expression: {}".format(repr(rule.expr_raw)),
                            level="ERROR",
                        )
                elif rule.value is not None:
                    plain_value = True
                    result = rule.value
                    log("=> {}".format(repr(result)), path, level="DEBUG")

                if isinstance(
                    result, expression.types.IncludeSchedule
                ) and path.includes_schedule(result.schedule):
                    # Prevent reusing IncludeSchedule results that would
                    # lead to a cycle. This happens when a rule of an
                    # included schedule returns Inherit() and the search
                    # then reaches the IncludeSchedule within the parent.
                    log(
                        "==   skipping in favour of parent to prevent cycle",
                        path,
                        level="DEBUG",
                    )
                    result = None
                elif result is None or isinstance(result, expression.types.Inherit):
                    log("==   skipping in favour of parent", path, level="DEBUG")
                    result = None
                else:
                    break

            if result is None:
                room.log(
                    "No expression/value definition found, skipping {}.".format(path),
                    level="WARNING",
                )
            elif isinstance(result, Exception):
                room.log(
                    "Evaluation failed, skipping {}.".format(path), level="WARNING"
                )
            elif isinstance(result, expression.types.Abort):
                break
            elif isinstance(result, expression.types.Break):
                prefix_size = max(0, len(path.rules) - result.levels)
                prefix = path.rules[:prefix_size]
                log("== breaking out of {}".format(prefix), path, level="DEBUG")
                while (
                    path_idx < len(paths)
                    and paths[path_idx].root_schedule == path.root_schedule
                    and paths[path_idx].rules[:prefix_size] == prefix
                ):
                    del paths[path_idx]
            elif isinstance(result, expression.types.IncludeSchedule):
                # Replace current rule with temporary SubScheduleRule to enable
                # proper handling of Inherit() and Break()
                _path = path.copy()
                _path.pop()
                _path.append(SubScheduleRule(result.schedule))
                paths.insert(path_idx, _path)
                for i, sub_path in enumerate(result.schedule.unfolded):
                    paths.insert(path_idx + i + 1, _path + sub_path)
            elif isinstance(result, expression.types.Postprocessor):
                if isinstance(result, expression.types.PostprocessorValueMixin):
                    value = room.validate_value(result.value)
                    if value is None:
                        room.log("Aborting schedule evaluation.", level="ERROR")
                        break
                    result.value = value
                postprocessors.append(result)
            elif isinstance(result, expression.types.Next):
                continue
            else:
                postprocessor_markers = set()  # type: T.Set[str]
                result = room.validate_value(result)
                if result is None and plain_value:
                    room.log(
                        "Maybe this is an expression? If so, set it "
                        "as the rule's 'expression' parameter "
                        "rather than as 'value'.",
                        level="WARNING",
                    )
                elif postprocessors:
                    room.log("Applying postprocessors.", level="DEBUG")
                    for postprocessor in postprocessors:
                        if result is None:
                            break
                        markers.update(postprocessor_markers)
                        postprocessor_markers.clear()
                        room.log("+ {}".format(repr(postprocessor)), level="DEBUG")
                        try:
                            result = postprocessor.apply(result)
                        except expression.types.PostprocessingError as err:
                            room.log(
                                "Error while applying {} to result {}: {}".format(
                                    repr(postprocessor), repr(result), err
                                ),
                                level="ERROR",
                            )
                            result = None
                            break
                        room.log("= {}".format(repr(result)), level="DEBUG")
                        if isinstance(result, expression.types.Mark):
                            result = result.unwrap(postprocessor_markers)
                        result = room.validate_value(result)

                if result is None:
                    room.log("Aborting schedule evaluation.", level="ERROR")
                    break
                markers.update(postprocessor_markers)

                room.log(
                    "Final result: {!r}, markers: {}".format(result, markers),
                    level="DEBUG",
                )
                return result, markers, last_rule

        room.log("Found no result.", level="DEBUG")
        return None

    def get_next_scheduling_datetime(
        self, now: datetime.datetime
    ) -> T.Optional[datetime.datetime]:
        """Returns a datetime object with the time at which the next
        re-scheduling should be done. now should be a datetime object
        containing the current date and time.
        SubScheduleRule objects and their rules are considered as well.
        None is returned in case there are no rules in the schedule."""

        times = self.get_scheduling_times()
        if not times:
            # no rules in schedule
            return None

        current_time = now.time()
        today = now.date()
        tomorrow = today + datetime.timedelta(days=1)

        def map_func(_time: datetime.time) -> datetime.datetime:
            """Maps a time object to a datetime containing the next
            occurrence of that time. Midnight transitions are handled
            correctly."""

            if _time <= current_time:
                # midnight transition
                return datetime.datetime.combine(tomorrow, _time)
            return datetime.datetime.combine(today, _time)

        return min(map(map_func, times))

    def get_scheduling_times(self) -> T.Set[datetime.time]:
        """Returns a set of times a re-scheduling should be triggered
        at. Rules of sub-schedules are considered as well."""

        times = set()  # type: T.Set[datetime.time]
        for path in self.unfolded:
            if not path.is_always_active:
                start_time, _, end_time, _ = path.times
                times.update((start_time, end_time))
        return times

    def unfolded_gen(self) -> T.Generator[RulePath, None, None]:
        """Implements recursive building of RulePath objects as a generator.
        It's like the unfolded property, but without being cached."""

        for rule in self.rules:
            path = RulePath(self)
            path.append(rule)
            yield path
            if isinstance(rule, SubScheduleRule):
                for sub_path in rule.sub_schedule.unfolded_gen():
                    yield path + sub_path

    @cached_property
    def unfolded(self) -> T.Tuple[RulePath, ...]:
        """Returns a tuple of rule paths.
        The last rule of a path may either be a SubScheduleRule (meaning
        the path leads to a node) or a Rule (meaning the path leads to
        a leaf). A node is returned first, followed by it's successors
        (like in depth-first search).
        NOTE: This is a cached property and only evaluated once."""

        return tuple(self.unfolded_gen())



"""
This module implements the Room class.
"""

import types
import typing as T

if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    import uuid
    from .app import SchedyApp
    from .schedule import Schedule
    from .actor.base import ActorBase

import datetime
import functools
import os
import threading
import traceback

from .. import common
from . import expression, util


def sync_proxy(handler: T.Callable) -> T.Callable:
    """A decorator for wrapping event and state handlers.
    It can be applied to members of Room or of objects having a Room
    object as their "room" attribute.
    It ensures all handlers are executed synchronously by acquiring a
    re-entrant lock stored in the Room object.
    Room._update_state() is called after the outmost handler wrapped
    with this decorator finishes."""

    @functools.wraps(handler)
    def wrapper(self: T.Any, *args: T.Any, **kwargs: T.Any) -> T.Any:
        """Wrapper around the event handler."""

        # pylint: disable=protected-access

        if isinstance(self, Room):
            room = self
        else:
            room = self.room

        with room._sync_proxy_lock:
            first = not room._sync_proxy_running
            try:
                if first:
                    room._sync_proxy_running = True
                result = handler(self, *args, **kwargs)
                if first:
                    room._update_state()
            finally:
                if first:
                    room._sync_proxy_running = False
            return result

    return wrapper


class Room:
    """A room to be controlled by Schedy."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, name: str, cfg: dict, app: "SchedyApp") -> None:
        self.name = name
        self.cfg = cfg
        self.app = app
        self.actors = []  # type: T.List[ActorBase]
        self.schedule = None  # type: T.Optional[Schedule]

        self._wanted_value = None  # type: T.Any
        self._scheduled_value = None  # type: T.Any
        self._rescheduling_time = None  # type: T.Optional[datetime.datetime]
        self._rescheduling_timer = None  # type: T.Optional[uuid.UUID]
        self._overlay_active = False
        self._overlaid_wanted_value = None  # type: T.Any
        self._overlaid_scheduled_value = None  # type: T.Any
        self._overlaid_rescheduling_time = None  # type: T.Optional[datetime.datetime]

        self._last_state = None  # type: T.Optional[T.Tuple[str, T.Dict[str, T.Any]]]

        self._sync_proxy_lock = threading.RLock()
        self._sync_proxy_running = False
        self._timer_lock = threading.RLock()

        self._reevaluation_timer = None  # type: T.Optional[uuid.UUID]

    def __repr__(self) -> str:
        return "<Room {}>".format(str(self))

    def __str__(self) -> str:
        return "R:{}".format(self.cfg.get("friendly_name", self.name))

    def _clear_overlay(self) -> None:
        """Removes all stored overlay state."""

        self._overlay_active = False
        self._overlaid_wanted_value = None
        self._overlaid_scheduled_value = None
        self._overlaid_rescheduling_time = None

    @sync_proxy
    def _initialize_actor_cb(self, kwargs: dict) -> None:
        """Is called for each actor until it's initialized successfully."""

        actor = kwargs["actor"]
        if not actor.initialize():
            self.log(
                "Actor {} couldn't be initialized, "
                "retrying in 10 seconds.".format(repr(actor)),
                level="WARNING",
            )
            self.app.run_in(self._initialize_actor_cb, 10, actor=actor)
            return

        actor.events.on("value_changed", self.notify_value_changed)
        if (
            self._wanted_value is not None
            and self.cfg["replicate_changes"]
            and all([a.is_initialized for a in self.actors])
        ):
            self.set_value(self._wanted_value)

    def _restore_state(self) -> None:
        """Restores a stored state from Home Assistant and.applies it.
        If no state was found, the schedule is just applied."""

        def _deserialize(value: T.Any) -> T.Any:
            if value is None:
                return None
            assert self.app.actor_type is not None
            return self.app.actor_type.deserialize_value(value)

        def _deserialize_dt(value: T.Any) -> T.Optional[datetime.datetime]:
            if not isinstance(value, (float, int)):
                return None
            return datetime.datetime.fromtimestamp(value)

        entity_id = self._state_entity_id
        self.log(
            "Loading state of {} from Home Assistant.".format(repr(entity_id)),
            level="DEBUG",
            prefix=common.LOG_PREFIX_OUTGOING,
        )
        state = self.app.get_state(entity_id, attribute="all")
        self.log(
            "  = {}".format(repr(state)),
            level="DEBUG",
            prefix=common.LOG_PREFIX_INCOMING,
        )

        reset = False
        if isinstance(state, dict):
            attrs = state.get("attributes", {})
            actor_wanted_values = attrs.get("actor_wanted_values", {})
            for entity_id, value in actor_wanted_values.items():
                value = _deserialize(value)
                for actor in self.actors:
                    if actor.entity_id == entity_id:
                        actor.wanted_value = value
            self._wanted_value = _deserialize(state.get("state") or None)
            self._scheduled_value = _deserialize(attrs.get("scheduled_value"))
            self._rescheduling_time = _deserialize_dt(attrs.get("rescheduling_time"))
            self._overlay_active = attrs.get("overlay_active") or False
            self._overlaid_wanted_value = _deserialize(
                attrs.get("overlaid_wanted_value")
            )
            self._overlaid_scheduled_value = _deserialize(
                attrs.get("overlaid_scheduled_value")
            )
            self._overlaid_rescheduling_time = _deserialize_dt(
                attrs.get("overlaid_rescheduling_time")
            )

            if self._rescheduling_time:
                if self._rescheduling_time > self.app.datetime():
                    if self.cfg["replicate_changes"]:
                        self.set_value(self._wanted_value)
                    self.start_rescheduling_timer(self._rescheduling_time)
                else:
                    self._rescheduling_time = None
                    reset = True

        self.apply_schedule(reset=reset)

    @sync_proxy
    def _rescheduling_timer_cb(self, kwargs: dict) -> None:
        """Is called whenever a re-scheduling timer fires.
        reset may be given in kwargs and is passed to apply_schedule()."""

        self.log("Re-scheduling timer fired.", level="DEBUG")
        with self._timer_lock:
            self._rescheduling_time = None
            self._rescheduling_timer = None
        self.apply_schedule(reset=True)

    @sync_proxy
    def _scheduling_timer_cb(self, kwargs: dict) -> None:
        """Is called whenever a scheduling timer fires."""

        self.log("Scheduling timer fired.", level="DEBUG")
        self.apply_schedule()

    @property
    def _state_entity_id(self) -> str:
        """Generates the entity id for storing this room's state as."""

        return "schedy_room.{}_{}".format(self.app.name, self.name)

    def _store_for_overlaying(self) -> None:
        """This method is called before a value overlay is put into place.
        It stores the scheduled and wanted value together with the re-scheduling
        time for later restoration. A running re-scheduling timer is cancelled.
        If an overlay is active already, this does nothing."""

        if self._overlay_active:
            # Don't overwrite existing restoration records
            return

        self.log("Storing value {!r} before overlaying.".format(self._wanted_value))
        self._overlay_active = True
        self._overlaid_wanted_value = self._wanted_value
        self._overlaid_scheduled_value = self._scheduled_value
        self._overlaid_rescheduling_time = self._rescheduling_time
        self.cancel_rescheduling_timer()

    def _update_state(self) -> None:
        """Update the room's state in Home Assistant."""

        D = T.TypeVar("D")

        def _serialize(value: T.Any, default: D) -> T.Union[str, D]:
            if value is None:
                return default
            assert self.app.actor_type is not None
            return self.app.actor_type.serialize_value(value)

        def _serialize_dt(value: T.Optional[datetime.datetime]) -> T.Optional[float]:
            if value is None:
                return None
            return value.timestamp()

        def _maybe_add(key: str, value: T.Any) -> None:
            if value is not None:
                attrs[key] = value

        state = _serialize(self._wanted_value, "")
        attrs = {
            "actor_wanted_values": {
                actor.entity_id: _serialize(actor.wanted_value, None)
                for actor in self.actors
            },
            "scheduled_value": _serialize(self._scheduled_value, None),
            "rescheduling_time": _serialize_dt(self._rescheduling_time),
            "overlay_active": self._overlay_active,
        }  # type: T.Dict[str, T.Any]
        _maybe_add(
            "overlaid_wanted_value", _serialize(self._overlaid_wanted_value, None)
        )
        _maybe_add(
            "overlaid_scheduled_value", _serialize(self._overlaid_scheduled_value, None)
        )
        _maybe_add(
            "overlaid_rescheduling_time",
            _serialize_dt(self._overlaid_rescheduling_time),
        )
        _maybe_add("friendly_name", self.cfg.get("friendly_name"))

        unchanged = (state, attrs) == self._last_state
        if unchanged:
            self.log(
                "Unchanged HA state: state={}, attributes={}".format(
                    repr(state), attrs
                ),
                level="DEBUG",
            )
            return
        self.log(
            "Sending state to HA: state={}, attributes={}".format(repr(state), attrs),
            level="DEBUG",
            prefix=common.LOG_PREFIX_OUTGOING,
        )

        entity_id = self._state_entity_id
        self.app.set_state(entity_id, state=state, attributes=attrs)
        self._last_state = (state, attrs)

    def validate_value(self, value: T.Any) -> T.Any:
        """A wrapper around self.app.actor_type.validate_value() that
        sanely logs validation errors and returns None in that case."""

        assert self.app.actor_type is not None
        try:
            value = self.app.actor_type.validate_value(value)
        except ValueError as err:
            self.log(
                "Invalid value {} for actor type {}: {}".format(
                    repr(value), repr(self.app.actor_type.name), err
                ),
                level="ERROR",
            )
            return None
        return value

    @sync_proxy
    def apply_schedule(self, reset: bool = False, force_resend: bool = False) -> None:
        """Applies the value scheduled for the current date and time.
        It detects when the result hasn't changed compared to the last
        run and prevent re-setting it in that case.
        This method will also not re-apply the schedule if a re-schedule
        timer runs - however, the OVERLAY marker is regarded.
        These both checks can be skipped by setting reset to True.
        force_resend is passed through to set_value()."""

        def _restore_overlaid_value() -> bool:
            """Restores and clears an overlaid value.
            Returns whether a value has actually been restored or not."""

            overlaid_wanted_value = self._overlaid_wanted_value
            if overlaid_wanted_value is None:
                self.log(
                    "Overlay ended but knowing no value to restore.", level="WARNING"
                )
                self._clear_overlay()
                return False

            delay = None  # type: T.Union[None, int, datetime.datetime]
            if not self._overlaid_rescheduling_time:
                if new_scheduled_value == self._overlaid_scheduled_value:
                    # Scheduled value hasn't changed compared to before overlay,
                    # hence revert to overlaid wanted value without timer
                    delay = 0
            elif self._overlaid_rescheduling_time > self.app.datetime():
                # Resume overlaid re-scheduling timer
                delay = self._overlaid_rescheduling_time
            else:
                self.log(
                    "Overlaid value {!r} has expired, not restoring it.".format(
                        overlaid_wanted_value
                    )
                )

            self._clear_overlay()
            if delay is None:
                return False
            self.log("Restoring overlaid value {}.".format(repr(overlaid_wanted_value)))
            self.set_value_manually(
                value=overlaid_wanted_value, rescheduling_delay=delay
            )
            return True

        self.log(
            "Evaluating room's schedule (reset={}, force_resend={}).".format(
                reset, force_resend
            ),
            level="DEBUG",
        )

        result = None
        if self.schedule:
            result = self.schedule.evaluate(self, self.app.datetime())
        if result is None:
            self.log("No suitable value found in schedule.", level="DEBUG")
            # revert an eventual overlay
            if self._overlay_active:
                new_scheduled_value = self._overlaid_scheduled_value
                _restore_overlaid_value()
                self._scheduled_value = new_scheduled_value
            return

        new_scheduled_value, markers = result[:2]
        if not (
            # Skip when scheduled value hasn't changed
            new_scheduled_value != self._scheduled_value
            # Never skip overlay activation
            or not self._overlay_active
            and expression.types.Mark.OVERLAY in markers
            # ... and deactivation
            or self._overlay_active
            and expression.types.Mark.OVERLAY not in markers
            # Don't skip if reset explicitly wanted
            or reset
            or force_resend
        ):
            self.log("Result didn't change, not setting it again.", level="DEBUG")
            return

        if reset:
            self.cancel_rescheduling_timer()
            self._clear_overlay()
        elif expression.types.Mark.OVERLAY in markers:
            # Create restoration records if overlay not active already
            self._store_for_overlaying()
        # No overlay should be set, hence try to revert an existing one
        elif self._overlay_active and _restore_overlaid_value():
            self._scheduled_value = new_scheduled_value
            return
        elif self._rescheduling_timer:
            self.log(
                "Not applying the schedule now due to a running "
                "re-scheduling timer.",
                level="DEBUG",
            )
            return

        self._scheduled_value = new_scheduled_value
        self.set_value(new_scheduled_value, force_resend=force_resend)

    def cancel_rescheduling_timer(self) -> bool:
        """Cancels the re-scheduling timer for this room, if one
        exists.
        Returns whether a timer has been cancelled."""

        with self._timer_lock:
            timer = self._rescheduling_timer
            if timer is None:
                return False
            self.app.cancel_timer(timer)
            self._rescheduling_time = None
            self._rescheduling_timer = None

        self.log("Cancelled re-scheduling timer.", level="DEBUG")
        return True

    def eval_expr(self, expr: types.CodeType, env: T.Dict[str, T.Any]) -> T.Any:
        """This is a wrapper around expression.eval_expr().
        It catches and logs any exception raised during evaluation. In
        this case, the caught Exception object is returned."""

        try:
            return expression.eval_expr(expr, env)
        except Exception as err:  # pylint: disable=broad-except
            self.log("Error while evaluating expression:", level="ERROR")
            tb_exc = traceback.TracebackException.from_exception(err)
            while tb_exc.stack and tb_exc.stack[0].filename != "expression":
                del tb_exc.stack[0]
            for line in tb_exc.format():
                self.log(line.rstrip(os.linesep), level="ERROR")
            return err

    @sync_proxy
    def initialize(self, reset: bool = False) -> None:
        """Should be called after all schedules and actors have been
        added in order to register state listeners and timers.
        If reset is True, the previous state won't be restored from Home
        Assistant and the schedule will be applied instead."""

        self.log("Initializing room (name={}).".format(repr(self.name)), level="DEBUG")

        for actor in self.actors:
            self._initialize_actor_cb({"actor": actor})

        assert self.schedule is not None
        times = self.schedule.get_scheduling_times()
        for snippet in self.app.cfg["schedule_snippets"].values():
            times.update(snippet.get_scheduling_times())
        self.log(
            "Registering scheduling timers at: [{}]".format(
                ", ".join(str(_time) for _time in sorted(times))
            ),
            level="DEBUG",
        )
        for _time in times:
            self.app.run_daily(self._scheduling_timer_cb, _time)

        if reset:
            self.apply_schedule(reset=True)
        else:
            self._restore_state()

        for definition in self.cfg["watched_entities"]:
            self.app.watch_entity(definition, [self])

    def log(self, msg: str, *args: T.Any, **kwargs: T.Any) -> None:
        """Prefixes the room to log messages."""

        msg = "[{}] {}".format(self, msg)
        self.app.log(msg, *args, **kwargs)

    @sync_proxy
    def notify_set_value_event(
        self,
        expr_raw: str = None,
        value: T.Any = None,
        force_resend: bool = False,
        rescheduling_delay: T.Union[float, int, None] = None,
    ) -> None:
        """Handles a schedy_set_value event for this room."""

        self.log(
            "schedy_set_value event received, [{}] "
            "[rescheduling_delay={}]".format(
                "expression={}".format(repr(expr_raw))
                if expr_raw is not None
                else "value={}".format(repr(value)),
                rescheduling_delay,
            ),
            prefix=common.LOG_PREFIX_INCOMING,
        )
        self.set_value_manually(
            expr_raw=expr_raw,
            value=value,
            force_resend=force_resend,
            rescheduling_delay=rescheduling_delay,
        )

    def notify_value_changed(self, actor: "ActorBase", value: T.Any) -> None:
        """Should be called when an actor reports a value change.
        Handles change replication and re-scheduling timers."""

        if actor.is_sending:
            self.log("Not respecting value change from a sending actor.", level="DEBUG")
            return
        if actor.gave_up_sending:
            self.log(
                "Not respecting value change from an actor that " "gave up sending.",
                level="DEBUG",
            )
            return

        actor_wanted = actor.wanted_value
        was_actor_wanted = actor_wanted is not None and value == actor_wanted
        replicating = self.cfg["replicate_changes"]
        single_actor = len([a.is_initialized for a in self.actors]) == 1

        if was_actor_wanted:
            synced = all(
                actor.is_initialized and actor.is_synced for actor in self.actors
            )
            if self.tracking_schedule and synced:
                self.cancel_rescheduling_timer()
        else:
            if not self.cfg["allow_manual_changes"]:
                if self._wanted_value is None:
                    self.log(
                        "Not rejecting manual value change by {} to "
                        "{} because we don't know what else to set.".format(
                            actor, repr(value)
                        )
                    )
                else:
                    self.log(
                        "Rejecting manual value change by {} to {}.".format(
                            actor, repr(value)
                        )
                    )
                    self.set_value(self._wanted_value)
                return

            if replicating:
                if not single_actor:
                    self.log("Propagating the change to all actors in the room.")
                self.set_value(value)

            if self.cfg["rescheduling_delay"] and not (
                replicating and self.tracking_schedule
            ):
                self.start_rescheduling_timer()

    def set_value(self, value: T.Any, force_resend: bool = False) -> None:
        """Sets the given value for all actors in the room.
        Values won't be send to actors redundantly unless force_resend
        is True."""

        scheduled = self._scheduled_value is not None and value == self._scheduled_value

        self.log(
            "Setting value to {}.  [{}{}]".format(
                repr(value),
                "scheduled" if scheduled else "manual",
                ", force re-sending" if force_resend else "",
            ),
            level="DEBUG",
        )

        self._wanted_value = value

        changed = False
        for actor in self.actors:
            if not actor.is_initialized:
                self.log(
                    "Skipping uninitialized {}.".format(repr(actor)), level="DEBUG"
                )
                continue
            changed |= actor.set_value(value, force_resend=force_resend)[0]

        if changed:
            self.log(
                "Value set to {}.  [{}]".format(
                    repr(value), "scheduled" if scheduled else "manual"
                ),
                prefix=common.LOG_PREFIX_OUTGOING,
            )

    def set_value_manually(
        self,
        expr_raw: str = None,
        value: T.Any = None,
        force_resend: bool = False,
        rescheduling_delay: T.Union[
            float, int, datetime.datetime, datetime.timedelta
        ] = None,
    ) -> None:
        """Evaluates the given expression or value and sets the result.
        An existing re-scheduling timer is cancelled and a new one is
        started if re-scheduling timers are configured.
        rescheduling_delay, if given, overwrites the value configured
        for the room. Passing 0 disables re-scheduling."""

        _checks = expr_raw is None, value is None
        if all(_checks) or not any(_checks):
            raise ValueError("specify exactly one of expr_raw and value")

        markers = set()  # type: T.Set[str]
        now = self.app.datetime()
        if expr_raw is not None:
            try:
                expr = util.compile_expression(expr_raw)
            except SyntaxError:
                for line in traceback.format_exc(limit=0):
                    self.log(line.rstrip(os.linesep), level="ERROR")
                self.log("Failed expression: {}".format(repr(expr_raw)), level="ERROR")
                return
            env = expression.build_expr_env(self, now)
            result = self.eval_expr(expr, env)
            self.log(
                "Evaluated expression {} to {}.".format(repr(expr_raw), repr(result)),
                level="DEBUG",
            )

            if isinstance(result, Exception):
                self.log("Failed expression: {}".format(repr(expr_raw)), level="ERROR")
                return

            if isinstance(result, expression.types.Mark):
                result = result.unwrap(markers)

            not_allowed_result_types = (
                expression.types.ControlResult,
                expression.types.Postprocessor,
                type(None),
            )
            value = None
            if isinstance(result, expression.types.IncludeSchedule):
                _result = result.schedule.evaluate(self, now)
                if _result is not None:
                    value = _result[0]
                    markers.update(_result[1])
            elif not isinstance(result, not_allowed_result_types):
                value = result

        if value is not None:
            value = self.validate_value(value)

        if value is None:
            self.log("Ignoring value.")
            return

        if expression.types.Mark.OVERLAY in markers:
            self._store_for_overlaying()

        self.set_value(value, force_resend=force_resend)
        if rescheduling_delay != 0:
            self.start_rescheduling_timer(delay=rescheduling_delay)
        else:
            self.cancel_rescheduling_timer()

    def start_rescheduling_timer(
        self, delay: T.Union[float, int, datetime.datetime, datetime.timedelta] = None
    ) -> None:
        """This method registers a re-scheduling timer according to the
        room's settings. delay, if given, overwrites the rescheduling_delay
        configured for the room. If there is a timer running already,
        it's replaced by a new one."""

        self.cancel_rescheduling_timer()

        if delay is None:
            delay = self.cfg["rescheduling_delay"]

        if isinstance(delay, (float, int)):
            delta = datetime.timedelta(minutes=delay)
            when = self.app.datetime() + delta
        elif isinstance(delay, datetime.datetime):
            delta = delay - self.app.datetime()
            when = delay
        elif isinstance(delay, datetime.timedelta):
            delta = delay
            when = self.app.datetime() + delay

        self.log(
            "Re-applying the schedule not before {} (in {}).".format(
                util.format_time(when.time()), delta
            )
        )

        with self._timer_lock:
            self._rescheduling_time = when
            self._rescheduling_timer = self.app.run_at(
                self._rescheduling_timer_cb, when
            )

    @property
    def tracking_schedule(self) -> bool:
        """Returns whether the value wanted by this room is the scheduled
        one."""

        return (
            self._scheduled_value is not None
            and self._wanted_value is not None
            and self._scheduled_value == self._wanted_value
        )

    def trigger_reevaluation(self, reset: bool = False) -> None:
        """Initializes a schedule re-evaluation in 1 second.
        The reset parameter is passed through to apply_schedule()."""

        def _reevaluation_cb(*args: T.Any, **kwargs: T.Any) -> None:
            with self._timer_lock:
                self._reevaluation_timer = None
            self.apply_schedule(reset=reset)

        with self._timer_lock:
            if self._reevaluation_timer:
                if reset:
                    self.app.cancel_timer(self._reevaluation_timer)
                else:
                    self.log("Re-evaluation pending, doing nothing.", level="DEBUG")
                    return

            self.log(
                "Doing schedule re-evaluation in 1 second [reset={}]".format(reset),
                level="DEBUG",
            )
            self._reevaluation_timer = self.app.run_in(_reevaluation_cb, 1)



"""
A multi-purpose scheduler for Home Assistant + AppDaemon.
"""

import typing as T

if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    import types
    from .actor.base import ActorBase
    from .room import Room
    from .stats import StatisticalParameter

from .. import common
from . import __version__, config, util


__all__ = ["SchedyApp"]


class SchedyApp(common.App):
    """The Schedy app class for AppDaemon."""

    class Meta(common.App.Meta):
        # pylint: disable=missing-docstring
        name = "schedy"
        version = __version__
        config_schema = config.CONFIG_SCHEMA

    def __init__(self, *args: T.Any, **kwargs: T.Any) -> None:
        super().__init__(*args, **kwargs)

        self.actor_type = None  # type: T.Optional[T.Type[ActorBase]]
        self.rooms = []  # type: T.List[Room]
        self.stats_params = []  # type: T.List[StatisticalParameter]
        self.expression_environment_script = None  # type: T.Optional[types.CodeType]

    def _check_accept_event(self, event: str, data: dict) -> bool:
        """Returns whether this Schedy instance is addressed by the
        given event name and data."""

        app_name = data.get("app_name", self.name)
        if app_name != self.name:
            self.log(
                "Ignoring {} event for app_name '{}', "
                "ours is '{}'.".format(event, app_name, self.name),
                level="DEBUG",
            )
            return False
        return True

    def _get_event_rooms(self, event: str, room_names: T.Any) -> T.Iterable["Room"]:
        """Returns an iterable over the rooms whose names were passed
        for the given event name. Passing None as room_names causes all
        rooms to be returned. A string or list of strings is converted
        to a list of the corresponding Room objects."""

        rooms = []
        if room_names is None:
            rooms.extend(self.rooms)
        else:
            if isinstance(room_names, str):
                room_names = [room_names]
            elif not isinstance(room_names, list):
                room_names = []
            for room_name in room_names:
                room = self.get_room(room_name)
                if room:
                    rooms.append(room)
                else:
                    self.log(
                        "Ignoring {} event for unknown room {}.".format(
                            event, repr(room_name)
                        ),
                        level="WARNING",
                    )

        return rooms

    def _reevaluate_event_cb(self, event: str, data: dict, kwargs: dict) -> None:
        """This callback executes when a schedy_reevaluate event is received.
        data may contain a "room_name", which limits the re-evaluation
        to the given room.
        mode has to be one of "reevaluate" (the default, and
        "reset"). It controls the value passed as the reset argument to
        Room.trigger_reevaluation()."""

        if not self._check_accept_event(event, data):
            return

        mode = data.get("mode", "reevaluate")
        if mode not in ("reset", "reevaluate"):
            self.log("Unknown mode {}.".format(repr(mode)), level="ERROR")
            return

        rooms = self._get_event_rooms(event, data.get("room"))
        self.log(
            "Re-evaluate event received for: {} [mode={}]".format(
                ", ".join([str(room) for room in rooms]), repr(mode)
            ),
            prefix=common.LOG_PREFIX_INCOMING,
        )

        for room in rooms:
            room.trigger_reevaluation(reset=bool(mode == "reset"))

    def _set_value_event_cb(self, event: str, data: dict, kwargs: dict) -> None:
        """This callback executes when a schedy_set_value event is received.
        data must contain a "room_name" and an "expression"/"x" or
        "value"/"v".
        "force_resend" is optional and False by default. If it is set
        to True, the value is re-sent to the actorss even if it hasn't
        changed."""

        if not self._check_accept_event(event, data):
            return

        try:
            rescheduling_delay = data.get("rescheduling_delay")
            if isinstance(rescheduling_delay, str):
                rescheduling_delay = float(rescheduling_delay)
            if not isinstance(rescheduling_delay, (type(None), float, int)):
                raise TypeError()
            if isinstance(rescheduling_delay, (float, int)) and rescheduling_delay < 0:
                raise ValueError()
            util.normalize_dict_key(data, "expression", "x")
            expr = data.get("expression")
            util.normalize_dict_key(data, "value", "v")
            value = data.get("value")
            if expr is not None and value is not None:
                raise ValueError()
            if expr is not None:
                if not self.cfg["expressions_from_events"]:
                    self.log(
                        "Received a {} event with an expression, "
                        "but expressions_from_events is not enabled "
                        "in your config. Ignoring event.".format(event),
                        level="ERROR",
                    )
                    raise ValueError()
            elif value is None:
                raise ValueError()
        except (TypeError, ValueError):
            self.log(
                "Ignoring {} event with invalid data: {}".format(event, repr(data)),
                level="WARNING",
            )
            return

        rooms = self._get_event_rooms(event, data.get("room"))
        for room in rooms:
            room.notify_set_value_event(
                expr_raw=expr,
                value=value,
                force_resend=bool(data.get("force_resend")),
                rescheduling_delay=rescheduling_delay,
            )

    def get_room(self, room_name: str) -> T.Optional["Room"]:
        """Returns the room with given name or None, if no such room
        exists."""

        for room in self.rooms:
            if room.name == room_name:
                return room
        return None

    def initialize_inner(self) -> None:
        """Initializes all actors and callbacks."""

        assert self.actor_type is not None
        self.log("Actor type is: {}".format(repr(self.actor_type.name)))

        if self.cfg["expression_environment"]:
            self.log("Compiling the expression_environment script.", level="DEBUG")
            self.expression_environment_script = compile(
                self.cfg["expression_environment"],
                "expression_environment",
                "exec",
                dont_inherit=True,
            )

        for room in self.rooms:
            room.initialize(reset=self.cfg["reset_at_startup"])

        for definition in self.cfg["watched_entities"]:
            self.watch_entity(definition, self.rooms)

        self.log("Listening for schedy_reevaluate event.", level="DEBUG")
        self.listen_event(self._reevaluate_event_cb, "schedy_reevaluate")

        self.log("Listening for schedy_set_value event.", level="DEBUG")
        self.listen_event(self._set_value_event_cb, "schedy_set_value")

        for stats_param in self.stats_params:
            stats_param.initialize()

    def watch_entity(
        self, definition: T.Dict[str, T.Any], rooms: T.List["Room"]
    ) -> None:
        """Sets up a state listener as configured in the given definition
        that triggers schedule re-evaluation in the given rooms."""

        def _cb(
            entity_id: str, attribute: str, old: T.Any, new: T.Any, kwargs: T.Any
        ) -> None:
            if new == old:
                return

            mode_str = "resetting" if reset else "reevaluating"
            if len(rooms) == 1:
                rooms_str = repr(rooms[0])
            else:
                rooms_str = "{} rooms".format(len(rooms))
            if attribute == "all":
                self.log(
                    "State of {} changed, {} {}.".format(
                        repr(entity_id), mode_str, rooms_str
                    ),
                    prefix=common.LOG_PREFIX_INCOMING,
                )
            else:
                self.log(
                    "Attribute {} of {} changed from {} to {}, {} {}.".format(
                        repr(attribute),
                        repr(entity_id),
                        repr(old),
                        repr(new),
                        mode_str,
                        rooms_str,
                    ),
                    prefix=common.LOG_PREFIX_INCOMING,
                )

            for room in rooms:
                room.trigger_reevaluation(reset=reset)

        entity_id = definition["entity"]
        attributes = definition["attributes"]
        mode = definition["mode"]

        self.log(
            "Watching {} for changes [attributes={}, rooms={}, mode={}]".format(
                repr(entity_id), attributes, rooms, mode
            ),
            level="DEBUG",
        )
        if mode != "ignore":
            reset = bool(mode == "reset")
            for attribute in attributes:
                self.listen_state(_cb, entity_id, attribute=attribute)



"""
Module containing types to be used during expression evaluation.
"""

import typing as T

if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    from .. import schedule


__all__ = [
    "Abort",
    "Break",
    "IncludeSchedule",
    "Inherit",
    "Mark",
    "Next",
    "Add",
    "Multiply",
    "Invert",
    "Postprocess",
]


class PostprocessingError(Exception):
    """Raised when Postprocessor.apply() fails."""


class PostprocessorValueMixin:
    """Makes a Postprocessor having a value to be validated by the used
    actor type."""

    def __init__(self, value: T.Any) -> None:
        self.value = value

    def __eq__(self, other: T.Any) -> bool:
        return super().__eq__(other) and self.value == other.value


class Postprocessor:
    """A postprocessor for the scheduling result."""

    def __eq__(self, other: T.Any) -> bool:
        return type(self) is type(other)

    def apply(self, result: T.Any) -> T.Any:
        """Implements the postprocessor-specific logic for updating the
        result. The returned value should have the type of result."""

        raise NotImplementedError()


class Add(PostprocessorValueMixin, Postprocessor):
    """Adds a value to the final result."""

    def apply(self, result: T.Any) -> T.Any:
        try:
            return type(result)(result + self.value)
        except TypeError as err:
            raise PostprocessingError(repr(err))

    def __repr__(self) -> str:
        return "Add({})".format(repr(self.value))


class Multiply(PostprocessorValueMixin, Postprocessor):
    """Multiplies a value with the final result."""

    def apply(self, result: T.Any) -> T.Any:
        try:
            return type(result)(result * self.value)
        except TypeError as err:
            raise PostprocessingError(repr(err))

    def __repr__(self) -> str:
        return "Multiply({})".format(repr(self.value))


class Invert(Postprocessor):
    """Negates the final result by calling __neg__().
    Booleans and the strings "on"/"off" are inverted instead."""

    specials = {True: False, False: True, "on": "off", "off": "on"}

    def apply(self, result: T.Any) -> T.Any:
        try:
            return self.specials[result]
        except KeyError:
            try:
                return -result
            except TypeError as err:
                raise PostprocessingError(repr(err))

    def __repr__(self) -> str:
        return "Invert()"


class Postprocess(Postprocessor):
    """A type which can be used for post-processing the later result by
    a custom function (i.e. a lambda closure)."""

    def __init__(self, func: T.Callable[[T.Any], T.Any]) -> None:
        self._func = func

    def apply(self, result: T.Any) -> T.Any:
        return self._func(result)


class ControlResult:
    """A special expression result used for evaluation flow control."""

    def __eq__(self, other: T.Any) -> bool:
        return type(self) is type(other)


class Abort(ControlResult):
    """Result of an expression that should cause scheduling to be aborted
    and the value left unchanged."""

    def __repr__(self) -> str:
        return "Abort()"


class Break(ControlResult):
    """Result of an expression that should cause the rest of a
    sub-schedule to be skipped."""

    def __init__(self, levels: int = 1) -> None:
        if not isinstance(levels, int) or levels < 1:
            raise ValueError(
                "levels to break must be >= 1, but is {}".format(repr(levels))
            )
        self.levels = levels

    def __eq__(self, other: T.Any) -> bool:
        return super().__eq__(other) and self.levels == other.levels

    def __repr__(self) -> str:
        return "Break({})".format(self.levels if self.levels != 1 else "")


class IncludeSchedule(ControlResult):
    """Result that inserts a schedule in place for further processing."""

    def __init__(self, sched: "schedule.Schedule") -> None:
        self.schedule = sched

    def __eq__(self, other: T.Any) -> bool:
        return super().__eq__(other) and self.schedule == other.schedule

    def __repr__(self) -> str:
        return "IncludeSchedule({})".format(self.schedule)


class Inherit(ControlResult):
    """Causes the next parent's value to be used as the result."""

    def __repr__(self) -> str:
        return "Inherit()"


class Mark(ControlResult):
    """A result with some markers applied."""

    # available markers
    OVERLAY = "OVERLAY"

    def __init__(self, result: T.Any, *markers: str) -> None:
        self.markers = set(markers)
        if isinstance(result, Mark):
            # Unwrap the result, combining present markers with our own ones
            self.result = result.unwrap(self.markers)
        else:
            self.result = result

    def __eq__(self, other: T.Any) -> bool:
        return (
            isinstance(other, self.__class__)
            and self.result == other.result
            and self.markers == other.markers
        )

    def __repr__(self) -> str:
        return "Mark({}, {})".format(repr(self.result), self.markers)

    def unwrap(self, markers_set: T.Set[str]) -> T.Any:
        """Returns the real, wrapped result. The applied markers will
        be added to markers_set."""

        markers_set.update(self.markers)
        return self.result


class Next(ControlResult):
    """Result of an expression which should be ignored."""

    def __repr__(self) -> str:
        return "Next()"



"""
Common helpers and functionality used by all apps.
"""

import typing as T

import copy

import voluptuous as vol
import voluptuous.humanize  # pylint: disable=unused-import

from appdaemon.plugins.hass import hassapi
from appdaemon.utils import __version__ as AD_VERSION


# prefixes for log messages
LOG_PREFIX_NONE = ""
LOG_PREFIX_STATUS = "---"
LOG_PREFIX_ALERT = "***"
LOG_PREFIX_WARNING = "!!!"
LOG_PREFIX_INCOMING = "-->"
LOG_PREFIX_OUTGOING = "<--"


class App(hassapi.Hass):  # type: ignore
    """
    This is a sub-class of hassapi.Hass which adds some common
    functionality. It's used by all apps included in hass_apps.
    """

    class Meta:
        """
        A class holding information about the app such as name and version.
        This information is used for logging, for instance.
        """

        name = "UNCONFIGURED"
        version = "0.0.0"
        config_schema = None  # type: T.Optional[T.Callable]

    def log(  # pylint: disable=arguments-differ
        self, msg: str, level: str = "INFO", prefix: T.Optional[str] = None
    ) -> None:
        """Wrapper around super().log() which changes the log level
        from DEBUG to INFO if debug config option is enabled.
        It also adds an appropriate prefix to the log message."""

        level = level.upper()
        if level == "DEBUG" and self.args and self.args.get("debug"):
            level = "INFO"

        if prefix is None:
            if level in ("DEBUG", "INFO"):
                prefix = LOG_PREFIX_STATUS
            elif level in ("WARNING", "ERROR"):
                prefix = LOG_PREFIX_WARNING

        if prefix:
            msg = "{} {}".format(prefix, msg)

        super().log(msg, level=level)

    def initialize(self) -> None:
        """Parses the configuration and logs that initialization
        started/finished. The real work should be done in
        initialize_inner()."""

        # pylint: disable=attribute-defined-outside-init

        def alert(*args: T.Any, **kwargs: T.Any) -> None:
            kwargs["prefix"] = LOG_PREFIX_ALERT
            self.log(*args, **kwargs)

        alert(
            "Welcome to {} {}, running on AppDaemon {}.".format(
                self.Meta.name, self.Meta.version, AD_VERSION
            )
        )
        alert("")
        alert("This is an app from the hass-apps package.")
        alert("  DOCS: https://hass-apps.readthedocs.io/en/stable/")
        alert("")
        alert("You like this app, want to honor the effort put into")
        alert("it, ensure continuous development and support?")
        alert("Then please consider making a donation.")
        alert("  DONATE: https://hass-apps.readthedocs.io/en/stable/#donations")
        alert("Thank you very much and enjoy {}!".format(self.Meta.name))
        alert("")

        if callable(self.Meta.config_schema):
            self.log("Validating the app's configuration.", level="DEBUG")
            cfg = copy.deepcopy(self.args)
            # Make the app object available during config validation.
            cfg["_app"] = self
            try:
                self.cfg = self.Meta.config_schema(cfg)  # pylint: disable=not-callable
            except vol.Invalid as err:
                msg = vol.humanize.humanize_error(cfg, err)
                self.log("Configuration error: {}".format(msg), level="ERROR")
                self.log("Not initializing this app.", level="ERROR")
                return

        self.initialize_inner()

        alert("Initialization done.")

    def initialize_inner(self) -> None:
        """Overwrite this stub to do the real initialization of the
        particular app."""
